Attribute VB_Name = "modTerrainCache"
Option Explicit
Option Base 1

Private Const GRIMBLE_ONLY As Boolean = False

'***************************************************************************
'** Publicly Exposed Constants                                            **
'***************************************************************************
Public Const TERRAIN_MODEL_NAME             As String = "terrain.rwx"

'***************************************************************************
'** Internal Constants                                                    **
'***************************************************************************
Public TERRAIN_WORLD_COORD_SIZE             As Long '= 3000 changed from constant
Public TERRAIN_DEFAULT_HEIGHT               As Long '= -10 changed from const

Private Const TERRAIN_COMMAND_PREFIX        As String = "d:"
Private Const TERRAIN_CELL_AREA_SIZE_MAX    As Long = 2
Private Const TERRAIN_QUERY_AREA_SIZE       As Long = 2
Private Const TERRAIN_QUERY_BUFFER_MAX      As Long = 10

Private Const TERRAIN_NO_CELL_INDEX         As Long = 0
Private Const TERRAIN_NO_VISIBLE_INDEX      As Long = 0

Private Const TERRAIN_NO_MARKER             As Long = 0
Private Const TERRAIN_TEXTURE_DEFAULT       As Long = 0

Private Const TERRAIN_MARKER_ACTION_SHOW    As String = vbNullString
Private Const TERRAIN_MARKER_ACTION_HIDE    As String = "create visible no"

Public Const TERRAIN_ENABLE_MESSAGE         As String = "Terrain features have been enabled for this session"
Public Const TERRAIN_DISABLE_MESSAGE        As String = "Terrain features have been disabled for this session"
Public Const TERRAIN_DISABLE_WARN_MESSAGE   As String = "Terrain features will be auto-disabled for this session"
Private Const TERRAIN_NOT_ENABLED_MESSAGE   As String = "Terrain features have not been enabled for this session"
Private Const TERRAIN_NO_ED_MESSAGE         As String = "You must have eminent domain in this world to access this function"

Private Const TERRAIN_SESSION_ALL           As Long = 0
Private Const TERRAIN_NO_SESSION            As Long = 0

Private Const TERRAIN_MARKER_CMD_SHOW       As String = "1"
Private Const TERRAIN_MARKER_CMD_HIDE       As String = "0"

'***************************************************************************
'** Data Structures For Terrain Cache                                     **
'***************************************************************************
Private Type STRUCT_TERRAIN_CACHE_X
  CoordX                    As Long
  ObjectNumber              As Long
  Texture                   As Long
  Height                    As Long
  Timestamp                 As Long
  Owner                     As Long
  Visible                   As Boolean
  RefreshTerrain            As Boolean
End Type

Private Type STRUCT_TERRAIN_CACHE_Z
  cellCount                 As Long
  XCells()                  As STRUCT_TERRAIN_CACHE_X
End Type

Private Type STRUCT_TERRAIN_MARKER
  Valid                     As Boolean
  CoordX                    As Long
  CoordZ                    As Long
  CellX                     As Long
  CellZ                     As Long
  Owner                     As Long
  Number                    As Long
  Height                    As Long
  Texture                   As Long
  IsHole                    As Boolean
  Rotation                  As AW_TERRAIN_ROTATE_TYPE
  Timestamp                 As Long
  Visible                   As Boolean
End Type

'***************************************************************************
'** Enums for Terrain Cache                                               **
'***************************************************************************
Private Enum TERRAIN_INVALID_ACTION_TYPE
  TERRAIN_INVALID_ACTION_DELETE
  TERRAIN_INVALID_ACTION_UPDATE
  TERRAIN_INVALID_ACTION_IGNORE
End Enum

'***************************************************************************
'** Show/Hide Background Queue Stores                                     **
'***************************************************************************
Private mShowHideQueue                      As New Collection

'***************************************************************************
'** Terrain Cache Stores                                                  **
'***************************************************************************
Private mActiveState                        As Boolean

Private mQueryBufferCount                   As Long
Private mCellQueryQueue                     As New Collection
Private mCellQueryX                         As Long
Private mCellQueryZ                         As Long
Private mCellQueryRefresh                   As Boolean

Private mTerrainCellCount                   As Long
Private mTerrainMarkerCount                 As Long
Private mTerrainCache()                     As STRUCT_TERRAIN_CACHE_Z

Public Sub TerrainCacheInitialize()
  '*************************************************************************
  '** Initialise all terrain stores and indicators                        **
  '*************************************************************************
  TerrainCacheInitStore
  TerrainCacheSetActiveState True, False
End Sub
'***************************************************************************
'** Initialise the terrain cache                                          **
'***************************************************************************
Public Sub TerrainEventSet()
  
  '*************************************************************************
  '** Ensure all required callbacks and event handlers are installed      **
  '*************************************************************************
  mAWSDK.aw_callback_set AW_CALLBACK_OBJECT_RESULT
  mAWSDK.aw_callback_set AW_CALLBACK_CELL_RESULT
  mAWSDK.aw_callback_set AW_CALLBACK_TERRAIN_SET_RESULT
  
  mAWSDK.aw_event_set AW_EVENT_OBJECT_ADD
  mAWSDK.aw_event_set AW_EVENT_OBJECT_DELETE
  mAWSDK.aw_event_set AW_EVENT_OBJECT_SELECT
  mAWSDK.aw_event_set AW_EVENT_CELL_OBJECT
  mAWSDK.aw_event_set AW_EVENT_CELL_BEGIN
  mAWSDK.aw_event_set AW_EVENT_CELL_END
  
  mAWSDK.aw_event_set AW_EVENT_CHAT
  mAWSDK.aw_event_set AW_EVENT_WORLD_DISCONNECT
  mAWSDK.aw_event_set AW_EVENT_WORLD_ATTRIBUTES
  
End Sub


'***************************************************************************
'** Terminate The Terrain Cache (Cleanup)                                 **
'***************************************************************************
Public Sub TerrainCacheTerm()

  '*************************************************************************
  '** Disable Terrain features                                            **
  '*************************************************************************
  TerrainCacheSetActiveState False, True
  
  '*************************************************************************
  '** Discard the terrain cache                                           **
  '*************************************************************************
  mTerrainCellCount = 0
  mTerrainMarkerCount = 0
  FormSetMarkerCount mTerrainMarkerCount
  Erase mTerrainCache
  
  mQueryBufferCount = 0
  
End Sub


'***************************************************************************
'** Initialise/Reset all terrain cache contents                           **
'***************************************************************************
Public Sub TerrainCacheInitStore()

  Dim cacheCellZ            As Long
  
  '*************************************************************************
  '** Initialise terrain cache store                                      **
  '*************************************************************************
  mTerrainCellCount = 0
  FormSetCellCount mTerrainCellCount
  
  mTerrainMarkerCount = 0
  FormSetMarkerCount mTerrainMarkerCount
  
  ReDim mTerrainCache(-TERRAIN_WORLD_COORD_SIZE To TERRAIN_WORLD_COORD_SIZE)
  
  For cacheCellZ = -TERRAIN_WORLD_COORD_SIZE To TERRAIN_WORLD_COORD_SIZE
    With mTerrainCache(cacheCellZ)
      .cellCount = 0
      Erase .XCells
      
    End With
  Next cacheCellZ

End Sub


'***************************************************************************
'** Handle The Deletion Of A Terrain Marker (Manual Or As Part Of Move)   **
'***************************************************************************
Private Sub TerrainCacheMarkerDelete(ByVal objCellX As Long, ByVal objCellZ As Long, ByVal objNumber As Long)

  Static cellXIndex         As Long
  Static avatarSession      As Long
  Static sessionEnabled     As Boolean
  Static terrainMarker      As STRUCT_TERRAIN_MARKER
  
  '*************************************************************************
  '** Locate that cell in the Terrain Cache                               **
  '*************************************************************************
  cellXIndex = TerrainCacheGetCellXIndex(objCellX, objCellZ)
    
  '*************************************************************************
  '** Determine whether the object was deleted by an enabled user         **
  '** the previously known information.                                   **
  '*************************************************************************
  avatarSession = mAWSDK.aw_int(AW_OBJECT_SESSION)
  
  If (avatarSession = mBotSession) Then
    sessionEnabled = True
    
  Else
    sessionEnabled = SessionTableGetSession(avatarSession).TerrainEnabled
    
  End If
        
  '*************************************************************************
  '** If the cell has no prior terrain information, query the cell        **
  '*************************************************************************
  If (cellXIndex = TERRAIN_NO_CELL_INDEX) Then
    TerrainCacheCellClear objCellX, objCellZ
    TerrainQueueAddCell objCellX, objCellZ
      
  '*************************************************************************
  '** If the cell is known, then there can only be one marker, so if the  **
  '** object number of the deleted object does not match the number held  **
  '** for the marker in the cell (may be zero) we can ignore the delete   **
  '*************************************************************************
  Else
    
    With mTerrainCache(objCellZ).XCells(cellXIndex)
      
      If (objNumber = .ObjectNumber) Then
        
        If Not sessionEnabled Then
          terrainMarker = TerrainCacheMarkerGet(objCellX, objCellZ)
          TerrainMarkerLoad terrainMarker
          TerrainCacheConsoleMessage avatarSession, TERRAIN_NOT_ENABLED_MESSAGE
          
        Else
        
          If Not (avatarSession = mBotSession) Then
            SessionTableSetTimer avatarSession, SESSION_TIMER_COMMAND
            
          End If
          
          '*******************************************************************
          '** We know about this cell, so we just invalidate the entry      **
          '*******************************************************************
          TerrainCacheCellClear objCellX, objCellZ
          
          '*******************************************************************
          '** Delete the terrain for the cell                               **
          '*******************************************************************
          TerrainCacheElevationSet objCellX, objCellZ, 0, TERRAIN_DEFAULT_HEIGHT
          
        End If
      End If
    End With
  End If

End Sub


'***************************************************************************
'***************************************************************************
'** SDK EVENT HANDLERS                                                    **
'***************************************************************************
'***************************************************************************

'***************************************************************************
'** Object Add Event                                                      **
'***************************************************************************
Public Sub TerrainEventObjectAdd()

  '*************************************************************************
  '** Handle any added terrain marker objects only                        **
  '*************************************************************************
  If Not (LCase(mAWSDK.aw_string(AW_OBJECT_MODEL)) = TERRAIN_MODEL_NAME) Then
    'Ignore non-terrain objects
    
  Else
    TerrainCacheMarkerAdd mAWSDK.aw_int(AW_OBJECT_SESSION), mAWSDK.aw_int(AW_OBJECT_X), mAWSDK.aw_int(AW_OBJECT_Y), mAWSDK.aw_int(AW_OBJECT_Z), mAWSDK.aw_int(AW_OBJECT_YAW), mAWSDK.aw_int(AW_OBJECT_TILT), mAWSDK.aw_int(AW_OBJECT_ROLL), _
                          mAWSDK.aw_int(AW_OBJECT_NUMBER), mAWSDK.aw_string(AW_OBJECT_DESCRIPTION), mAWSDK.aw_string(AW_OBJECT_ACTION), mAWSDK.aw_int(AW_OBJECT_OWNER), mAWSDK.aw_int(AW_OBJECT_BUILD_TIMESTAMP), _
                          True, TERRAIN_INVALID_ACTION_UPDATE
  
  End If
  
End Sub


'***************************************************************************
'** Object Delete Event                                                   **
'***************************************************************************
Public Sub TerrainEventObjectDelete()
  
  '*************************************************************************
  '** Handle the deletion of an object in the world                       **
  '*************************************************************************
  TerrainCacheMarkerDelete mAWSDK.aw_int(AW_CELL_X), mAWSDK.aw_int(AW_CELL_Z), mAWSDK.aw_int(AW_OBJECT_NUMBER)
  
End Sub


'***************************************************************************
'** Object Select Event                                                   **
'***************************************************************************
Public Sub TerrainEventObjectSelect()

  Dim objCellX              As Long
  Dim objCellZ              As Long
  
  objCellX = mAWSDK.aw_int(AW_CELL_X)
  objCellZ = mAWSDK.aw_int(AW_CELL_Z)
  
  If (TerrainCacheGetCellXIndex(objCellX, objCellZ) = TERRAIN_NO_CELL_INDEX) Then
    TerrainQueueAddCell objCellX, objCellZ
    
  End If
  
End Sub


'***************************************************************************
'** Query Cell Object Event                                               **
'***************************************************************************
Public Sub TerrainEventCellObject()

  '*************************************************************************
  '** Tell the cache to handle the new terrain model object               **
  '*************************************************************************
  If (LCase(mAWSDK.aw_string(AW_OBJECT_MODEL)) = TERRAIN_MODEL_NAME) Then
    TerrainCacheMarkerAdd mAWSDK.aw_int(AW_OBJECT_SESSION), mAWSDK.aw_int(AW_OBJECT_X), mAWSDK.aw_int(AW_OBJECT_Y), mAWSDK.aw_int(AW_OBJECT_Z), mAWSDK.aw_int(AW_OBJECT_YAW), mAWSDK.aw_int(AW_OBJECT_TILT), mAWSDK.aw_int(AW_OBJECT_ROLL), _
                          mAWSDK.aw_int(AW_OBJECT_NUMBER), mAWSDK.aw_string(AW_OBJECT_DESCRIPTION), mAWSDK.aw_string(AW_OBJECT_ACTION), mAWSDK.aw_int(AW_OBJECT_OWNER), mAWSDK.aw_int(AW_OBJECT_BUILD_TIMESTAMP), _
                          False, TERRAIN_INVALID_ACTION_UPDATE
    
  End If
  
  
End Sub


'***************************************************************************
'** Query Cell Begin Event                                                **
'***************************************************************************
Public Sub TerrainEventCellBegin()
          
  '*************************************************************************
  '** Update the global values for the current query cell                 **
  '*************************************************************************
  mCellQueryX = mAWSDK.aw_int(AW_CELL_X)
  mCellQueryZ = mAWSDK.aw_int(AW_CELL_Z)
  mCellQueryRefresh = False
  
End Sub


'***************************************************************************
'** Query Cell End Event                                                  **
'***************************************************************************
Public Sub TerrainEventCellEnd()

  Static cellXIndex         As Long
  
  '*************************************************************************
  '** Check the state of the query cell. If no object is known, delete    **
  '** the terrain, otherwise do nothing (already handled)
  '*************************************************************************
  cellXIndex = TerrainCacheGetCellXIndex(mCellQueryX, mCellQueryZ)
  If (cellXIndex = TERRAIN_NO_CELL_INDEX) Then
    TerrainCacheCellClear mCellQueryX, mCellQueryZ
  
  ElseIf mCellQueryRefresh Then
    With mTerrainCache(mCellQueryZ).XCells(cellXIndex)
      TerrainCacheElevationSet mCellQueryX, mCellQueryZ, .Texture, .Height
    
    End With
  End If
  
End Sub


'***************************************************************************
'** Chat Event                                                            **
'***************************************************************************
Public Sub TerrainEventChat()

  Static chatMessage        As String
  Static chatSession        As Long
  Static authorisedAvatar   As Boolean
  Static terrainCommand     As String
  
  'If the speaker doesn't have build, ignore all chat by exiting sub
  If Not HasRight(SessionTableGetSession(mAWSDK.aw_int(AW_CHAT_SESSION)).PrivNum, AW_WORLD_BUILD_RIGHT) Then Exit Sub
  
  
  chatMessage = LCase(mAWSDK.aw_string(AW_CHAT_MESSAGE))
  
  '*************************************************************************
  '** Ignore the chat if it is not a terrain command "terrain:" prefix    **
  '*************************************************************************
  If (Left(chatMessage, Len(TERRAIN_COMMAND_PREFIX)) = TERRAIN_COMMAND_PREFIX) Then
    chatSession = mAWSDK.aw_int(AW_CHAT_SESSION)
    
    '***********************************************************************
    '** Identify permitted "Super Users"                                  **
    '***********************************************************************
    authorisedAvatar = ((LCase(mAWSDK.aw_string(AW_AVATAR_NAME)) = "grimble") Or _
                        (LCase(mAWSDK.aw_string(AW_AVATAR_NAME)) = "brant"))
                      
    '***********************************************************************
    '** Handle command accordingly                                        **
    '***********************************************************************
    terrainCommand = Trim(Mid(chatMessage, Len(TERRAIN_COMMAND_PREFIX) + 1))
    Select Case terrainCommand
      
      '*********************************************************************
      '** Emergency shutdown - THIS DESTROYS THE INSTANCE IMMEDIATELY.    **
      '** Only available to "Super Users".                                **
      '*********************************************************************
      Case "version"
        TerrainCacheConsoleMessage chatSession, "This world is running Demeter" + Str$(App.Major) + "." + its(App.Minor) + its(App.Revision)
      
      Case "panic"
        If authorisedAvatar Then
          mAWSDK.aw_destroy
          FormSetState FORM_STATE_TERMINATED
          FormSetWorldName vbNullString
          FormSetConnectState FORM_CONNECTION_OFFLINE
          FormSetAvatarCount 0
          FormSetTerrainSessionCount 0
          TerrainCacheTerm
          
        End If
        
      '*********************************************************************
      '** Display the current stats of the cache                          **
      '*********************************************************************
      Case "stats"
          SessionTableSetTimer chatSession, SESSION_TIMER_COMMAND
          TerrainCacheConsoleMessage mAWSDK.aw_int(AW_CHAT_SESSION), TerrainCacheGetStats()
        
      '*********************************************************************
      '** Start terrain tracking for the user                             **
      '*********************************************************************
      Case "begin"
        If GRIMBLE_ONLY And SessionTableGetSession(chatSession).name <> "Grimble" Then
          TerrainCacheConsoleMessage chatSession, "Offline for testing"
          Exit Sub
        End If
        
        If SessionTableEnableTerrain(chatSession, True) Then
          TerrainCacheConsoleMessage chatSession, TERRAIN_ENABLE_MESSAGE
          
          '*****************************************************************
          '** Update the session timer for the last command               **
          '*****************************************************************
          SessionTableSetTimer chatSession, SESSION_TIMER_COMMAND
          
        End If
        
      '*********************************************************************
      '** Stop terrain tracking for the user                              **
      '*********************************************************************
      Case "end"
        If SessionTableEnableTerrain(chatSession, False) Then
          TerrainCacheConsoleMessage chatSession, TERRAIN_DISABLE_MESSAGE
          
        End If
        
      '*********************************************************************
      '** Show/hide know markers in avatar vicinity                       **
      '*********************************************************************
      Case "show", "hide", "show all", "hide all", "hide ed", "show ed"
        With SessionTableGetSession(chatSession)
          If Not .TerrainEnabled Then
            TerrainCacheConsoleMessage chatSession, TERRAIN_NOT_ENABLED_MESSAGE
          
          ElseIf Not HasRight(.PrivNum, AW_WORLD_EMINENT_DOMAIN_RIGHT) And _
            Right(chatMessage, 2) = "ed" Then
           'Check to see whether this user has ED and is trying to use an ED-only command
            TerrainCacheConsoleMessage chatSession, TERRAIN_NO_ED_MESSAGE
          
          Else
            Dim showMarkers     As Boolean
            Dim handleAll       As Boolean
            Dim queuedCount     As Long
        
            '***************************************************************
            '** Show or hide the markers as required                      **
            '***************************************************************
            showMarkers = (Left(terrainCommand, 4) = "show")
            
            If (Right(terrainCommand, 3) = "all") Then
              queuedCount = TerrainCacheMarkersShow(0, 0, _
                                        .PrivNum, showMarkers, TERRAIN_WORLD_COORD_SIZE)
            
            ElseIf (Right(terrainCommand, 2) = "ed") Then
              queuedCount = TerrainCacheMarkersShow(AWCellFromCoord(.X, AW_CELL_CALC_CELL), _
                                        AWCellFromCoord(.Z, AW_CELL_CALC_CELL), _
                                        .PrivNum, showMarkers, TERRAIN_CELL_AREA_SIZE_MAX, True)
              'The true argument at the end specifies to ignore the privilege number of the issuer
              'because this is an ED command
            Else
              queuedCount = TerrainCacheMarkersShow(AWCellFromCoord(.X, AW_CELL_CALC_CELL), _
                                        AWCellFromCoord(.Z, AW_CELL_CALC_CELL), _
                                        .PrivNum, showMarkers, TERRAIN_CELL_AREA_SIZE_MAX)
            End If
            
            '***************************************************************
            '** Update the session for the last command and msg the user  **
            '***************************************************************
            SessionTableSetTimer chatSession, SESSION_TIMER_COMMAND
            TerrainCacheConsoleMessage chatSession, queuedCount & " cells will be " _
                                                  & IIf(showMarkers, "displayed", "hidden")
          End If
        End With
          
      '*********************************************************************
      '** Seed                                                            **
      '*********************************************************************
      Case "seed"
        Dim seedSession     As STRUCT_AW_SESSION
        Dim newMarker       As STRUCT_TERRAIN_MARKER
        Dim seedCoordX      As Long
        Dim seedCoordZ      As Long
        Dim cellXIndex      As Long
        
        '*******************************************************************
        '** Deny action if terrain features are not enabled               **
        '*******************************************************************
        seedSession = SessionTableGetSession(chatSession)
        If Not seedSession.TerrainEnabled Then
          TerrainCacheConsoleMessage chatSession, TERRAIN_NOT_ENABLED_MESSAGE
            
        Else
            
          '*****************************************************************
          '** Create a marker from the avatar's current position          **
          '*****************************************************************
          With seedSession
            newMarker = TerrainCacheMarkerDefine(.X, .Y, .Z, TERRAIN_NO_MARKER, vbNullString, vbNullString, .PrivNum, 0)
            
          End With
          
          With newMarker
            '***************************************************************
            '** Deny the action if the avatar is outside the world limits **
            '***************************************************************
            If ((Abs(.CellX) > TERRAIN_WORLD_COORD_SIZE) Or (Abs(.CellZ) > TERRAIN_WORLD_COORD_SIZE)) Then
              TerrainCacheConsoleMessage chatSession, "You cannot seed an object outside the world building limits"

            Else
              
              '*************************************************************
              '** Define the new marker position at the cell vertex       **
              '*************************************************************
              .CoordX = .CellX * 1000
              .CoordZ = .CellZ * 1000
            
              '*************************************************************
              '** If the cache has not caught up with the avatar, deny    **
              '** the action with appropriate warning                     **
              '*************************************************************
              cellXIndex = TerrainCacheGetCellXIndex(.CellX, .CellZ)
              If (cellXIndex = TERRAIN_NO_CELL_INDEX) Then
                FormWriteToLog "Seed request failed for '" & seedSession.name & "' - Unknown Cell"
                TerrainCacheConsoleMessage chatSession, "Seed failed due to internal caching. Please try seeding again"
                
              Else
                '***********************************************************
                '** If the new marker will encroach on an existing one    **
                '** owned by another user, deny the action                **
                '***********************************************************
                If (Not (mTerrainCache(.CellZ).XCells(cellXIndex).Owner = seedSession.PrivNum) And _
                    Not (mTerrainCache(.CellZ).XCells(cellXIndex).ObjectNumber = TERRAIN_NO_MARKER)) Then
                  TerrainCacheConsoleMessage chatSession, "You cannot seed a new terrain marker in place of someone else's"
                
                Else
                  '*********************************************************
                  '** Load the new terrain marker                         **
                  '*********************************************************
                  TerrainMarkerLoad newMarker
                  TerrainCacheConsoleMessage chatSession, "New terrain marker seeded at " & AWLocationNarrative(.CoordX, .CoordZ)
                
                End If
              End If
            End If
          End With
        
          '*****************************************************************
          '** Update the session for the last command and msg the user    **
          '*****************************************************************
          SessionTableSetTimer chatSession, SESSION_TIMER_COMMAND
        End If
          
      '*********************************************************************
      '** Invalid Terrain Command                                         **
      '*********************************************************************
      Case Else
        TerrainCacheConsoleMessage chatSession, "'" & terrainCommand & "' is an invalid command"
        
    End Select
  End If
  
End Sub


'***************************************************************************
'** World Disconnect Event                                                **
'***************************************************************************
Public Sub TerrainEventWorldDisconnect()

  '*************************************************************************
  '** Log the disconnect event and update the form                        **
  '*************************************************************************
  FormWriteToLog "Disconnected from world"
  FormSetWorldName vbNullString
  
  'Clear the terrain cache if disconnected to eliminate the problem caused
  'by duplicate avatar entries upon reconnect
  SessionTableTerm
 
  'Update form to notify user of disconnect
  FormSetState FORM_STATE_DISCONNECTED
  FormSetConnectState FORM_CONNECTION_DISCONNECTED
  
End Sub


'***************************************************************************
'** World Attributes Event                                                **
'***************************************************************************
Public Sub TerrainEventWorldAttributes()

  '*************************************************************************
  '** New connection established - update connection stores and form      **
  '*************************************************************************
  mBotSession = mAWSDK.aw_session()
  
  FormSetWorldName mAWSDK.aw_string(AW_WORLD_NAME)
  
  'determine when to display the "World Connect" message
  Select Case FORM_CURRENT_STATE
    Case FORM_STATE_LOGGING_IN
      FormWriteToLog "Connected to world"
    Case FORM_STATE_DISCONNECTED
      FormWriteToLog "Reconnected to world"
  End Select
  
  'if not a caretaker, exit
  If mAWSDK.aw_bool(AW_WORLD_CARETAKER_CAPABILITY) <> 1 Then
    LogoutWithError "User does not have Caretaker right in this world"
    Exit Sub
  End If
 
  'if the program's logging in, then Demeter needs to get the world size before it can
  'get the marker count to avoid subscript out of range errors
  
  If FORM_CURRENT_STATE <> FORM_STATE_LOGGING_IN Then
    TerrainWorldAttributesGetMarkers
  End If
  
  FormSetState FORM_STATE_ACTIVE
  FormSetConnectState FORM_CONNECTION_ONLINE
  
End Sub

Public Sub TerrainWorldAttributesGetMarkers()
    
  'gets updates the marker count upon initial connect or upon reconnect
  mTerrainMarkerCount = TerrainCacheGetCount
  FormSetMarkerCount mTerrainMarkerCount
  
  mQueryBufferCount = 0

End Sub
'***************************************************************************
'** Object Result Callback                                                **
'***************************************************************************
Public Sub TerrainCallbackObjectResult(ByVal awRC As Long)

  'Do Nothing
  
End Sub


'***************************************************************************
'** Query Cell Result Callback                                            **
'***************************************************************************
Public Sub TerrainCallbackCellResult(ByVal awRC As Long)

  '*************************************************************************
  '** Attempt to process the next item in the query queue                 **
  '*************************************************************************
  If (mQueryBufferCount > 0) Then mQueryBufferCount = mQueryBufferCount - 1
  TerrainQueryQueueProcess
  
End Sub


'***************************************************************************
'** Query Terrain Set Callback                                            **
'***************************************************************************
Public Sub TerrainCallbackTerrainSetResult(ByVal awRC As Long)

  'Do Nothing
  
End Sub


'***************************************************************************
'** Handle the addition of a new qualifying object to the world           **
'***************************************************************************
Private Sub TerrainCacheMarkerAdd(ByVal objSession As Long, ByVal objX As Long, ByVal objY As Long, ByVal objZ As Long, ByVal objYaw As Long, ByVal objTilt As Long, ByVal objRoll As Long, ByVal objNumber As Long, ByVal objDescription As String, ByVal objAction As String, ByVal objOwner As Long, ByVal objTimestamp As Long, Optional updateTerrain As Boolean = True, Optional invalidAction As TERRAIN_INVALID_ACTION_TYPE = TERRAIN_INVALID_ACTION_UPDATE)

  Static thisMarker         As STRUCT_TERRAIN_MARKER
  Static fixedMarker        As STRUCT_TERRAIN_MARKER
  Static currMarker         As STRUCT_TERRAIN_MARKER
  Static cellXIndex         As Long
  Static avatarSession      As Long
    
  '*************************************************************************
  '** If the object was not placed by the bot or found through a query,   **
  '** delete the object if the session has not enabled terrain            **
  '*************************************************************************
  avatarSession = mAWSDK.aw_int(AW_OBJECT_SESSION)
    
  'WHY object_session?  |
    
  If Not (avatarSession = mBotSession) Then
    If updateTerrain Then
      SessionTableSetTimer avatarSession, SESSION_TIMER_COMMAND
      
      If Not SessionTableGetSession(avatarSession).TerrainEnabled Then
        mAWSDK.aw_object_delete
        TerrainCacheConsoleMessage avatarSession, TERRAIN_NOT_ENABLED_MESSAGE
        Exit Sub
        
      End If
    End If
  End If
  
  '*************************************************************************
  '** Create the marker details from the supplied attributes              **
  '*************************************************************************
  thisMarker = TerrainCacheMarkerDefine(objX, objY, objZ, objNumber, objDescription, objAction, objOwner, objTimestamp)
  
  '*************************************************************************
  '** If the marker does not comply to the standard form, update it       **
  '*************************************************************************
  If Not TerrainMarkerCorrect(thisMarker, fixedMarker, objYaw, objTilt, objRoll) Then
    If invalidAction Then
      TerrainMarkerLoad fixedMarker
      TerrainMarkerDelete thisMarker
      
    End If
  
  Else
    
    '***********************************************************************
    '** Attempt to locate the cell in the terrain cache                   **
    '***********************************************************************
    cellXIndex = TerrainCacheGetCellXIndex(thisMarker.CellX, thisMarker.CellZ)

    '***********************************************************************
    '** If the cell was not previously known, add it and queue a query    **
    '***********************************************************************
    If (cellXIndex = TERRAIN_NO_CELL_INDEX) Then
      cellXIndex = TerrainCacheCellSet(thisMarker, cellXIndex)
      TerrainQueueAddCell thisMarker.CellX, thisMarker.CellZ
      
      If updateTerrain Then
        TerrainMarkerElevationSet thisMarker
        
      End If
      
    Else
      '*********************************************************************
      '** Retrieve the marker from the cell in the terrain cache          **
      '*********************************************************************
      currMarker = TerrainCacheMarkerGet(thisMarker.CellX, thisMarker.CellZ, cellXIndex)
      
      '*********************************************************************
      '** If it is the same marker as before, so nothing. This is needed  **
      '** for when we use the AW_CELL_COMBINE = TRUE attribute on the     **
      '** cell queries, which gives us much better "read-ahead" even      **
      '** though the same cell may be notified several times              **
      '*********************************************************************
      If Not (currMarker.Number = thisMarker.Number) Then
        
        '*******************************************************************
        '** Compare the new marker to the current marker to see if an     **
        '** update is required                                            **
        '*******************************************************************
        If TerrainMarkerIsActive(currMarker, thisMarker) Then
          TerrainMarkerDelete currMarker
          TerrainCacheCellSet thisMarker, cellXIndex
          
          '*****************************************************************
          '** If we're not updating the terrain now, tell the query that  **
          '** it will need to at the end of this cell                     **
          '*****************************************************************
          If Not updateTerrain Then
            mCellQueryRefresh = True
            
          Else
            '***************************************************************
            '** Update the terrain if required                            **
            '***************************************************************
            TerrainMarkerElevationSet thisMarker
            
          End If
        
        Else
          '*****************************************************************
          '** Delete the current marker if it has been superceeded        **
          '*****************************************************************
          TerrainMarkerDelete thisMarker
        
        End If
      End If
    End If
  End If
    
End Sub

'***************************************************************************
'** Define a marker from raw object attributes                            **
'***************************************************************************
Private Function TerrainCacheMarkerDefine(ByVal objX As Long, ByVal objY As Long, ByVal objZ As Long, ByVal objNumber As Long, ByVal objDescription As String, ByVal objAction As String, ByVal objOwner As Long, ByVal objTimestamp As Long) As STRUCT_TERRAIN_MARKER

  Dim terrainMarker         As STRUCT_TERRAIN_MARKER
  
  '*************************************************************************
  '** Populate a new terrain marker from the supplied attributes          **
  '*********************************************************************bn****
  With terrainMarker
    .CoordX = objX
    .CoordZ = objZ
    .CellX = AWCellFromCoord(objX, AW_CELL_CALC_VERTEX)
    .CellZ = AWCellFromCoord(objZ, AW_CELL_CALC_VERTEX)
    .Height = objY
    .Number = objNumber
    .Timestamp = objTimestamp
    .Owner = objOwner
    .Visible = Not (objAction = TERRAIN_MARKER_ACTION_HIDE)
    
    .Valid = TerrainCacheExtractTexture(objDescription, .Texture, .Rotation, .IsHole)
    .Valid = .Valid And ((objAction = TERRAIN_MARKER_ACTION_HIDE) Or (objAction = TERRAIN_MARKER_ACTION_SHOW))
    .Valid = .Valid And (.CoordX = (.CellX * 1000))
    .Valid = .Valid And (.CoordZ = (.CellZ * 1000))
    
  End With
    
  TerrainCacheMarkerDefine = terrainMarker
  
End Function


'***************************************************************************
'** Build correctly populated marker, returning validity of the original  **
'***************************************************************************
Private Function TerrainMarkerCorrect(ByRef srcMarker As STRUCT_TERRAIN_MARKER, ByRef validMarker As STRUCT_TERRAIN_MARKER, ByVal objYaw As Long, ByVal objTilt As Long, ByVal objRoll As Long) As Boolean

  Dim markerValid           As Boolean
  
  With srcMarker
    
    '***********************************************************************
    '** Initialise the the vailidity from the current marker state        **
    '***********************************************************************
    markerValid = .Valid
    
    '***********************************************************************
    '** Copy over the static and confirmed attributes                     **
    '***********************************************************************
    validMarker.Texture = .Texture
    validMarker.IsHole = .IsHole
    validMarker.Rotation = .Rotation
    
    validMarker.CellX = .CellX
    validMarker.CellZ = .CellZ
    validMarker.CoordX = .CellX * 1000
    validMarker.CoordZ = .CellZ * 1000
    validMarker.Height = .Height
    
    validMarker.Visible = .Visible
    validMarker.Owner = .Owner
    validMarker.Number = 0
    validMarker.Timestamp = 0
    
    '***********************************************************************
    '** Check th orientation of the marker for zero values                **
    '***********************************************************************
    If Not ((objYaw = 0) And (objTilt = 0) And (objRoll = 0)) Then
      markerValid = False
      
    End If
    
    validMarker.Valid = True
    
  End With
  
  TerrainMarkerCorrect = markerValid
  
End Function


'***************************************************************************
'** Clear the appropriate entry in the cache, creating one if required    **
'***************************************************************************
Private Sub TerrainCacheCellClear(ByVal terrainCellX As Long, ByVal terrainCellZ As Long)

  Dim cellXIndex            As Long
  
  '*************************************************************************
  '** Attempt to locate the cell in the terrain cache                     **
  '*************************************************************************
  cellXIndex = TerrainCacheGetCellXIndex(terrainCellX, terrainCellZ)
  
  With mTerrainCache(terrainCellZ)
    
    '***********************************************************************
    '** Create a new cell in the terrain cache if one does not exist      **
    '***********************************************************************
    If (cellXIndex = TERRAIN_NO_CELL_INDEX) Then
      cellXIndex = TerrainCacheAddNewCell(terrainCellX, terrainCellZ)
      
    '***********************************************************************
    '** If the cell was not already empty, reduce the marker count        **
    '***********************************************************************
    ElseIf Not (.XCells(cellXIndex).ObjectNumber = TERRAIN_NO_MARKER) Then
      mTerrainMarkerCount = mTerrainMarkerCount - 1
      FormSetMarkerCount mTerrainMarkerCount
      
    End If
  
    '***********************************************************************
    '** Clear the cell                                                    **
    '***********************************************************************
    With .XCells(cellXIndex)
      .CoordX = terrainCellX
      .ObjectNumber = TERRAIN_NO_MARKER
      .Owner = TERRAIN_NO_MARKER
      .Timestamp = TERRAIN_NO_MARKER
      
    End With
  End With
  
End Sub


'***************************************************************************
'** Set the contents of a cell in the terrain cache                       **
'***************************************************************************
Private Function TerrainCacheCellSet(ByRef terrainMarker As STRUCT_TERRAIN_MARKER, ByVal cellXIndex As Long) As Long
  
  With mTerrainCache(terrainMarker.CellZ)
    
    '***********************************************************************
    '** If the cell was not found in the terrain cache, create a new one  **
    '***********************************************************************
    If (cellXIndex = TERRAIN_NO_CELL_INDEX) Then
      cellXIndex = TerrainCacheAddNewCell(terrainMarker.CellX, terrainMarker.CellZ)
      
      '*********************************************************************
      '** Increase the terrain marker count                               **
      '*********************************************************************
      mTerrainMarkerCount = mTerrainMarkerCount + 1
      FormSetMarkerCount mTerrainMarkerCount
    
    '***********************************************************************
    '** If the cell was previously know but clear, increase the count     **
    '***********************************************************************
    ElseIf (.XCells(cellXIndex).ObjectNumber = TERRAIN_NO_MARKER) Then
      mTerrainMarkerCount = mTerrainMarkerCount + 1
      FormSetMarkerCount mTerrainMarkerCount
      
    End If
  
    '***********************************************************************
    '** Update the terrain cache from the terrain marker supplied         **
    '***********************************************************************
    With .XCells(cellXIndex)
      .CoordX = terrainMarker.CellX
      .ObjectNumber = terrainMarker.Number
      .Texture = AWTerrainBuildTexture(terrainMarker.IsHole, terrainMarker.Texture, terrainMarker.Rotation)
      .Height = terrainMarker.Height
      .Owner = terrainMarker.Owner
      .Timestamp = terrainMarker.Timestamp
      .Visible = terrainMarker.Visible
      
    End With
  End With
  
  TerrainCacheCellSet = cellXIndex
  
End Function


'***************************************************************************
'** Delete a terrain marker from the world                                **
'***************************************************************************
Private Function TerrainMarkerDelete(ByRef terrainMarker As STRUCT_TERRAIN_MARKER) As Long

  Dim awRC                  As Long
  
  '*************************************************************************
  '** Set the SDK attributes for the marker to delete                     **
  '*************************************************************************
  With terrainMarker
    mAWSDK.aw_int_set AW_OBJECT_X, .CoordX
    mAWSDK.aw_int_set AW_OBJECT_Z, .CoordZ
    mAWSDK.aw_int_set AW_OBJECT_NUMBER, .Number
    
  End With
  
  '*************************************************************************
  '** Delete the marker from the world                                    **
  '*************************************************************************
  awRC = mAWSDK.aw_object_delete()
  If Not (awRC = AW_RC_OK) Then
    ErrorCode awRC
    FormWriteToLog "[ERROR]: AwObjectDelete (Reason = " & ErrReturn$ & ")"
    
  End If
  
  TerrainMarkerDelete = awRC
  
End Function


'***************************************************************************
'** Load a new terrain marker to the world                                **
'***************************************************************************
Private Function TerrainMarkerLoad(ByRef terrainMarker As STRUCT_TERRAIN_MARKER) As Long

  Dim awRC                  As Long
  
  With terrainMarker
    mAWSDK.aw_string_set AW_OBJECT_MODEL, TERRAIN_MODEL_NAME
    mAWSDK.aw_int_set AW_OBJECT_X, .CoordX
    mAWSDK.aw_int_set AW_OBJECT_Y, .Height
    mAWSDK.aw_int_set AW_OBJECT_Z, .CoordZ
    mAWSDK.aw_int_set AW_OBJECT_YAW, 0
    mAWSDK.aw_int_set AW_OBJECT_TILT, 0
    mAWSDK.aw_int_set AW_OBJECT_ROLL, 0
    mAWSDK.aw_string_set AW_OBJECT_DESCRIPTION, TerrainCacheBuildObjectDescription(terrainMarker)
    mAWSDK.aw_string_set AW_OBJECT_ACTION, IIf(.Visible, TERRAIN_MARKER_ACTION_SHOW, TERRAIN_MARKER_ACTION_HIDE)
    mAWSDK.aw_int_set AW_OBJECT_OWNER, .Owner
  
    awRC = mAWSDK.aw_object_load
    If Not (awRC = AW_RC_OK) Then
      ErrorCode awRC
      FormWriteToLog "[ERROR]: AwObjectLoad (Reason = " & ErrReturn$ & ")"
      
    End If
  End With
  
End Function


'***************************************************************************
'** Update the world terrain from a marker details                        **
'***************************************************************************
Private Function TerrainMarkerElevationSet(ByRef terrainMarker As STRUCT_TERRAIN_MARKER) As Long

  With terrainMarker
    TerrainMarkerElevationSet = TerrainCacheElevationSet(.CellX, .CellZ, _
                                AWTerrainBuildTexture(.IsHole, .Texture, .Rotation), _
                                 .Height)

  End With
  
End Function


'***************************************************************************
'** Update the world terrain for a given cell                             **
'***************************************************************************
Public Function TerrainCacheElevationSet(ByVal terrainCellX As Long, ByVal terrainCellZ, ByVal terrainTexture As Long, ByVal terrainHeight As Integer) As Long

  Dim awRC                  As Long

  mAWSDK.aw_terrain_heights(0) = terrainHeight
  awRC = mAWSDK.aw_terrain_set(terrainCellX, terrainCellZ, 1, terrainTexture)
  
  If Not (awRC = AW_RC_OK) Then
    ErrorCode awRC
    FormWriteToLog "[ERROR]: AwTerrainSet (Reason = " & ErrReturn$ & ")"
    
  End If
  
  TerrainCacheElevationSet = awRC

End Function


'***************************************************************************
'** Retrieve a terrain marker from the terrain cache                      **
'***************************************************************************
Public Function TerrainCacheMarkerGet(ByVal terrainCellX As Long, ByVal terrainCellZ As Long, Optional knownCellXIndex As Long = TERRAIN_NO_CELL_INDEX) As STRUCT_TERRAIN_MARKER

  Static cellXIndex         As Long
  Static terrainMarker      As STRUCT_TERRAIN_MARKER
  
  '*************************************************************************
  '** Set the non-cache related information                               **
  '*************************************************************************
  terrainMarker.CellX = terrainCellX
  terrainMarker.CellZ = terrainCellZ
  terrainMarker.CoordX = terrainCellX * 1000
  terrainMarker.CoordZ = terrainCellZ * 1000
  
  '*************************************************************************
  '** Attempt to retrieve the cell from the terrain cache                 **
  '*************************************************************************
  If (knownCellXIndex = TERRAIN_NO_CELL_INDEX) Then
    cellXIndex = TerrainCacheGetCellXIndex(terrainCellX, terrainCellZ)
  
  Else
    cellXIndex = knownCellXIndex
  
  End If
  
  '*************************************************************************
  '** If no cell was found, default the marker values                     **
  '*************************************************************************
  If (cellXIndex = TERRAIN_NO_CELL_INDEX) Then
    terrainMarker.Valid = False
    
    terrainMarker.Owner = TERRAIN_NO_MARKER
    terrainMarker.Number = TERRAIN_NO_MARKER
    terrainMarker.Height = TERRAIN_NO_MARKER
    terrainMarker.Timestamp = TERRAIN_NO_MARKER
    terrainMarker.Visible = True
    
    terrainMarker.IsHole = False
    terrainMarker.Rotation = AW_TERRAIN_ROTATE_NONE
    terrainMarker.Texture = TERRAIN_TEXTURE_DEFAULT
    
  Else
    
    '***********************************************************************
    '** Copy the cache cell to the marker                                 **
    '***********************************************************************
    With mTerrainCache(terrainCellZ).XCells(cellXIndex)
      terrainMarker.Valid = True
      
      terrainMarker.Owner = .Owner
      terrainMarker.Number = .ObjectNumber
      terrainMarker.Height = .Height
      terrainMarker.Timestamp = .Timestamp
      terrainMarker.Visible = .Visible
      
      '*********************************************************************
      '** If the texture is for a hole                                    **
      '*********************************************************************
      If (.Texture = AW_TERRAIN_TEXTURE_HOLE) Then
        terrainMarker.IsHole = True
        terrainMarker.Rotation = AW_TERRAIN_ROTATE_NONE
        terrainMarker.Texture = TERRAIN_TEXTURE_DEFAULT
        
      Else
        '*******************************************************************
        '** Extract the texture index from the texture value              **
        '*******************************************************************
        terrainMarker.IsHole = False
        terrainMarker.Texture = (.Texture And &H3F)
    
        '*******************************************************************
        '** Extract the rotation value from the texture value             **
        '*******************************************************************
        If ((.Texture And AW_TERRAIN_ROTATE_RIGHT) = AW_TERRAIN_ROTATE_RIGHT) Then
          terrainMarker.Rotation = AW_TERRAIN_ROTATE_RIGHT
          
        ElseIf ((.Texture And AW_TERRAIN_ROTATE_LEFT) = AW_TERRAIN_ROTATE_LEFT) Then
          terrainMarker.Rotation = AW_TERRAIN_ROTATE_LEFT
      
        ElseIf ((.Texture And AW_TERRAIN_ROTATE_OPPOSITE) = AW_TERRAIN_ROTATE_OPPOSITE) Then
          terrainMarker.Rotation = AW_TERRAIN_ROTATE_OPPOSITE
    
        Else
          terrainMarker.Rotation = AW_TERRAIN_ROTATE_NONE
          
        End If
      End If
    End With
  End If
  
  TerrainCacheMarkerGet = terrainMarker
    
End Function


'***************************************************************************
'** Determine whether a marker should become the active marker            **
'***************************************************************************
Private Function TerrainMarkerIsActive(ByRef currMarker As STRUCT_TERRAIN_MARKER, ByRef newMarker As STRUCT_TERRAIN_MARKER) As Boolean

  Dim newMarkerActive       As Boolean
  
  '*****************************************************************
  '** NOTE: Since all numerics in VB are signed, it is feasible   **
  '** that the timestamp could be negative for a large value. Due **
  '** to this, we need to bring the sign of the timestamp into    **
  '** into the comparison, making it messy and over-complicated.  **
  '*****************************************************************
  
  '*****************************************************************
  '** Both timestamps are greater than zero                       **
  '*****************************************************************
  If ((newMarker.Timestamp > 0) And (currMarker.Timestamp > 0)) Then
    newMarkerActive = (newMarker.Timestamp > currMarker.Timestamp)
    
  '*****************************************************************
  '** Both timestamps are less than zero                          **
  '*****************************************************************
  ElseIf ((newMarker.Timestamp < 0) And (currMarker.Timestamp < 0)) Then
    newMarkerActive = (newMarker.Timestamp < currMarker.Timestamp)
  
  '*****************************************************************
  '** New timestamp is GT zero, current timestamp is LT zero      **
  '*****************************************************************
  ElseIf ((newMarker.Timestamp > 0) And (currMarker.Timestamp < 0)) Then
    newMarkerActive = False
  
  '*****************************************************************
  '** Current timestamp is GT zero, new timestamp is LT zero      **
  '*****************************************************************
  Else
    newMarkerActive = True

  End If
  
  TerrainMarkerIsActive = newMarkerActive

End Function


'***************************************************************************
'** Extract texture attributes from the marker description text           **
'***************************************************************************
Private Function TerrainCacheExtractTexture(ByVal objDescription As String, ByRef textureIndex As Long, ByRef textureRotate As AW_TERRAIN_ROTATE_TYPE, ByRef textureHole As Boolean) As Boolean

  Dim markerValid           As Boolean
  Dim objTexture            As String
  Dim objRotation           As String
  Dim commaPos              As Long
  
  markerValid = True
  
  '*************************************************************************
  '** Clean up the object description text for analysis                   **
  '*************************************************************************
  objDescription = LCase(Trim(objDescription))
  
  '*************************************************************************
  '** Check for a hole (not valid with any other settings)                **
  '*************************************************************************
  If (objDescription = "hole") Then
    textureIndex = 0
    textureRotate = AW_TERRAIN_ROTATE_NONE
    textureHole = True
    
  Else
    
    textureHole = False
    
    '***********************************************************************
    '** Locate a comma delimiter. If not present, handle as rotation only **
    '***********************************************************************
    commaPos = InStr(objDescription, ",")
    If (commaPos = 0) Then
      textureRotate = AW_TERRAIN_ROTATE_NONE
      
      '*********************************************************************
      '** Invalid value - set to defaults                                 **
      '*********************************************************************
      If Not (CStr(Val(objDescription)) = objDescription) Then
        textureIndex = TERRAIN_TEXTURE_DEFAULT
        markerValid = False
        
      Else
        '*******************************************************************
        '** Extract the stated texture index                              **
        '*******************************************************************
        textureIndex = Val(objDescription)
        
      End If
    
      objRotation = "0"
      
    Else
      
      '*********************************************************************
      '** Separate the two components of the texture information          **
      '*********************************************************************
      objTexture = Trim(Left(objDescription, commaPos - 1))
      objRotation = Trim(Mid(objDescription, commaPos + 1))
      
      '*********************************************************************
      '** Interpret the texture index                                     **
      '*********************************************************************
      If (CStr(Val(objTexture)) = objTexture) Then
        textureIndex = Val(objTexture)
        
      Else
        textureIndex = TERRAIN_TEXTURE_DEFAULT
        markerValid = False
        
      End If
      
      '*********************************************************************
      '** Interpret the texture rotation                                  **
      '*********************************************************************
      Select Case objRotation
        Case "0"
          textureRotate = AW_TERRAIN_ROTATE_NONE
          
        Case "90"
          textureRotate = AW_TERRAIN_ROTATE_LEFT
          
        Case "180"
          textureRotate = AW_TERRAIN_ROTATE_OPPOSITE
          
        Case "270"
          textureRotate = AW_TERRAIN_ROTATE_RIGHT
          
        Case Else
          textureRotate = AW_TERRAIN_ROTATE_NONE
          objRotation = "0"
          markerValid = False
          
      End Select
      
    End If
  
    '***********************************************************************
    '** Ensure the texture index is within the valid range                **
    '***********************************************************************
    If Not ((textureIndex >= AW_TERRAIN_TEXTURE_MIN) And (textureIndex <= AW_TERRAIN_TEXTURE_MAX)) Then
      textureIndex = TERRAIN_TEXTURE_DEFAULT
      markerValid = False
      
    End If
    
    '***********************************************************************
    '** Reset the object description incase it needs updating             **
    '***********************************************************************
    objDescription = CStr(textureIndex) & "," & objRotation
      
  End If
  
  '*************************************************************************
  '** Return by indicating whether the texture information was valid      **
  '*************************************************************************
  TerrainCacheExtractTexture = markerValid

End Function


'***************************************************************************
'** Issue a console message to a specific session                         **
'***************************************************************************
Public Sub TerrainCacheConsoleMessage(ByVal msgSession As Long, ByVal msgText As String)

  mAWSDK.aw_int_set AW_CONSOLE_RED, 0
  mAWSDK.aw_int_set AW_CONSOLE_GREEN, 0
  mAWSDK.aw_int_set AW_CONSOLE_BLUE, 255
  mAWSDK.aw_bool_set AW_CONSOLE_ITALICS, False
  mAWSDK.aw_bool_set AW_CONSOLE_BOLD, False
  mAWSDK.aw_string_set AW_CONSOLE_MESSAGE, "Demeter:" & vbTab & msgText
  mAWSDK.aw_console_msg msgSession
  
End Sub


'***************************************************************************
'** Set the state of the terrain features and inform the users            **
'***************************************************************************
Public Sub TerrainCacheSetActiveState(ByVal newState As Boolean, ByVal announceState As Boolean)

  '*************************************************************************
  '** Ignore the new state if there is no change                          **
  '*************************************************************************
  If Not (mActiveState = newState) Then
    mActiveState = newState
    
    If announceState Then
      SessionStateMessageAll 'notify everyone with build terrain is going offline
      FormWriteToLog "Terrain Manager: " & IIf(newState, "ONLINE", "OFFLINE")
      
    End If
  End If
  
End Sub


'***************************************************************************
'** Issue State message to supplied session                               **
'***************************************************************************
Public Sub TerrainCacheStateMessage(ByVal msgSession As Long)

  TerrainCacheConsoleMessage msgSession, "Terrain features are " & IIf(mActiveState, "ONLINE", "OFFLINE")
  
End Sub


'***************************************************************************
'** Determine cache statistic from the store                              **
'***************************************************************************
Public Function TerrainCacheGetStats() As String

  Dim terrainCount    As Long
  Dim allCellCount    As Long
  Dim ZCountX         As Long
  Dim minCellX        As Long
  Dim maxCellX        As Long
  Dim minCellZ        As Long
  Dim maxCellZ        As Long
  Dim cellXIndex      As Long
  Dim CellZ           As Long
  
  '*************************************************************************
  '** Initialise counters and stores                                      **
  '*************************************************************************
  terrainCount = 0
  allCellCount = 0
  
  minCellX = TERRAIN_WORLD_COORD_SIZE
  maxCellX = -TERRAIN_WORLD_COORD_SIZE
  minCellZ = TERRAIN_WORLD_COORD_SIZE
  maxCellZ = -TERRAIN_WORLD_COORD_SIZE
  
  '*************************************************************************
  '** Scan the entire cache                                               **
  '*************************************************************************
  For CellZ = -TERRAIN_WORLD_COORD_SIZE To TERRAIN_WORLD_COORD_SIZE
    With mTerrainCache(CellZ)
      If (.cellCount > 0) Then
        
        ZCountX = 0
        
        For cellXIndex = 1 To .cellCount
          allCellCount = allCellCount + 1
          
          With .XCells(cellXIndex)
            
            '***************************************************************
            '** Ignore cleared cells                                      **
            '***************************************************************
            If Not (.ObjectNumber = TERRAIN_NO_MARKER) Then
              
              '*************************************************************
              '** Increment the X counter and maintain min/max X values   **
              '*************************************************************
              ZCountX = ZCountX + 1
              
              If (.CoordX < minCellX) Then minCellX = .CoordX
              If (.CoordX > maxCellX) Then maxCellX = .CoordX
            
            End If
          End With
        Next cellXIndex
          
        '*******************************************************************
        '** If populated cells were found, increment the Z counter and    **
        '** maintain min/max Z values
        '*******************************************************************
        If Not (ZCountX = 0) Then
          terrainCount = terrainCount + ZCountX
          If (CellZ < minCellZ) Then minCellZ = CellZ
          maxCellZ = CellZ
        
        End If
      End If
    End With
  Next CellZ
  
  '*************************************************************************
  '** Return the statistics information in formatted text                 **
  '*************************************************************************
  TerrainCacheGetStats = "Markers: " & terrainCount _
                       & ", Cells: " & allCellCount _
                       & IIf(terrainCount = 0, vbNullString, ", " _
                       & "Marker Extents: " & AWLocationNarrative(minCellX * 1000, minCellZ * 1000) _
                                   & " To " & AWLocationNarrative(maxCellX * 1000, maxCellZ * 1000))

End Function



'***************************************************************************
'** Add a new cell reference to the query queue                           **
'***************************************************************************
Public Sub TerrainQueueAddCell(ByVal terrainCellX As Long, ByVal terrainCellZ As Long)

  Static cellXIndex         As Long
  Static queueCell          As Boolean
  Static cellIterator       As Long
  
  '*************************************************************************
  '** Attempt to locate the cell in the terrain cache                     **
  '*************************************************************************
  cellXIndex = TerrainCacheGetCellXIndex(terrainCellX, terrainCellZ)
'  If (cellXIndex = TERRAIN_NO_CELL_INDEX) Then
    
    '***********************************************************************
    '** Queue the cell for querying (ignore error if already queued)      **
    '***********************************************************************
    On Error Resume Next
    cellIterator = AWBuildCellIterator(terrainCellX, terrainCellZ)
    mCellQueryQueue.Add cellIterator, CStr(cellIterator)
    
'  End If
  
End Sub


'***************************************************************************
'** Add an area of cells based on a centered box to the query queue       **
'***************************************************************************
Public Sub TerrainQueueAddArea(ByVal centerCellX As Long, ByVal centerCellZ As Long)

  Static queueCellX         As Long
  Static queueCellZ         As Long
  
  '*************************************************************************
  '** Attempt to queue each area cell that is within the world limits     **
  '*************************************************************************
  For queueCellZ = centerCellZ - TERRAIN_QUERY_AREA_SIZE To centerCellZ + TERRAIN_QUERY_AREA_SIZE
    If (Abs(queueCellZ) <= TERRAIN_WORLD_COORD_SIZE) Then
      For queueCellX = centerCellX - TERRAIN_QUERY_AREA_SIZE To centerCellX + TERRAIN_QUERY_AREA_SIZE
        If (Abs(queueCellX) <= TERRAIN_WORLD_COORD_SIZE) Then
          TerrainQueueAddCell queueCellX, queueCellZ
            
        End If
      Next queueCellX
    End If
  Next queueCellZ
  
End Sub


'***************************************************************************
'** Initiate the loop for a new queued query request                      **
'***************************************************************************
Public Sub TerrainQueryQueueProcess()

  Dim cellIterator          As String
  Dim queryCellX            As Long
  Dim queryCellZ            As Long
    
  '*************************************************************************
  '** Quick exits depending on request origin and queue state             **
  '*************************************************************************
  If (mCellQueryQueue.Count = 0) Then Exit Sub

  '*************************************************************************
  '** Keep sending queries until either the queue is empty or the maximum **
  '** number of queries are in progress                                   **
  '*************************************************************************
  Do While Not ((mCellQueryQueue.Count = 0) Or (mQueryBufferCount >= TERRAIN_QUERY_BUFFER_MAX))

    '***********************************************************************
    '** Decode the cell iterator for later reference                      **
    '***********************************************************************
    cellIterator = Right("00000000" & Hex(mCellQueryQueue(1)), 8)
    queryCellX = Val("&h" & Left(cellIterator, 4))
    queryCellZ = Val("&h" & Right(cellIterator, 4))
  
    '***********************************************************************
    '** If the cell is previously unknown, query it                       **
    '***********************************************************************
    If (TerrainCacheGetCellXIndex(queryCellX, queryCellZ) = TERRAIN_NO_CELL_INDEX) Then
      
      '*********************************************************************
      '** Issue a cell query for the first cell in the queue              **
      '*********************************************************************
      mQueryBufferCount = mQueryBufferCount + 1
      mAWSDK.aw_int_set AW_CELL_ITERATOR, mCellQueryQueue(1)
      mAWSDK.aw_bool_set AW_CELL_COMBINE, False
      mAWSDK.aw_cell_next
      
    End If
  
    '***********************************************************************
    '** Remove the cell from the queue, whether it was queried or not     **
    '***********************************************************************
    mCellQueryQueue.Remove (1)
      
  Loop
  
End Sub


'***************************************************************************
'** Show or hide a  set of terrain markers based on a centered box        **
'***************************************************************************
Public Function TerrainCacheMarkersShow(ByVal centerCellX As Long, ByVal centerCellZ As Long, ByVal filterCitNum As Long, ByVal showMarker As Boolean, ByVal boxSize As Long, Optional IgnoreCitnum As Boolean = False)

  Dim showCellZ             As Long
  Dim cellXIndex            As Long
  Dim cellMarker            As STRUCT_TERRAIN_MARKER
  Dim queuedCount           As Long
  
  queuedCount = 0
  
  '*************************************************************************
  '** Handle every area cell within the world limits                      **
  '*************************************************************************
  For showCellZ = centerCellZ - boxSize To centerCellZ + boxSize
    If (Abs(showCellZ) <= TERRAIN_WORLD_COORD_SIZE) Then
    
      With mTerrainCache(showCellZ)
        For cellXIndex = 1 To .cellCount
          With mTerrainCache(showCellZ).XCells(cellXIndex)
          
            If ((.CoordX >= centerCellX - boxSize) And _
                (.CoordX <= centerCellX + boxSize)) Then
            
              '*************************************************************
              '** If the cell is not clear and the marker belongs to the  **
              '** supplied citizen number, queue a show/hide request      **
              '*************************************************************
              'IgnoreCitnum tells the routine to ignore the citizen number so that
              'ED rightholders can control terrain vandalism
              If (((.Owner = filterCitNum) Or IgnoreCitnum) And Not (.ObjectNumber = TERRAIN_NO_MARKER)) Then
                If Not (.Visible = showMarker) Then
                  QueueShowHideRequest .CoordX, showCellZ, showMarker
                  queuedCount = queuedCount + 1
                  
                End If
              
              End If
            End If
          End With
        Next cellXIndex
      End With
    End If
  Next showCellZ
  
  TerrainCacheMarkersShow = queuedCount
  
End Function


'***************************************************************************
'** Retrieve the correct marker count from the entire cache               **
'***************************************************************************
Private Function TerrainCacheGetCount() As Long

  Dim terrainCellZ          As Long
  Dim terrainXIndex         As Long
  Dim terrainCount          As Long
  
  terrainCount = 0
  
  For terrainCellZ = -TERRAIN_WORLD_COORD_SIZE To TERRAIN_WORLD_COORD_SIZE
    With mTerrainCache(terrainCellZ)
      For terrainXIndex = 1 To .cellCount
        If Not (.XCells(terrainXIndex).ObjectNumber = TERRAIN_NO_MARKER) Then
          terrainCount = terrainCount + 1
          
        End If
      Next terrainXIndex
    End With
  Next terrainCellZ
  
  
  TerrainCacheGetCount = terrainCount

End Function

'***************************************************************************
'** Build a terrainmarker object description from texture attributes      **
'***************************************************************************
Private Function TerrainCacheBuildObjectDescription(ByRef terrainMarker As STRUCT_TERRAIN_MARKER)

  Dim objDescription        As String
  
  With terrainMarker
    If .IsHole Then
      objDescription = "hole"
      
    Else
      objDescription = .Texture
      
      Select Case .Rotation
        Case AW_TERRAIN_ROTATE_LEFT
          objDescription = objDescription & ",90"
          
        Case AW_TERRAIN_ROTATE_OPPOSITE
          objDescription = objDescription & ",180"
          
        Case AW_TERRAIN_ROTATE_RIGHT
          objDescription = objDescription & ",270"
          
        Case Else
          'Omit rotation if none specified
          
      End Select
    End If
  End With

  TerrainCacheBuildObjectDescription = objDescription
  
End Function


'***************************************************************************
'** ... and here we go with the crumby debug tools                        **
'***************************************************************************
Public Sub TerrainCacheReportMarkers()

  Dim CellZ                 As Long
  Dim indexX                As Long
  Dim terrainCount          As Long
  Dim cellCount             As Long
  Dim cellData              As String
  
  terrainCount = 0
  cellCount = 0
  
  FormWriteToLog "## Terrain Report ##"
  
  For CellZ = -TERRAIN_WORLD_COORD_SIZE To TERRAIN_WORLD_COORD_SIZE
    With mTerrainCache(CellZ)
      For indexX = 1 To .cellCount
        With .XCells(indexX)
          cellCount = cellCount + 1
          
          If Not (.ObjectNumber = TERRAIN_NO_MARKER) Then
            cellData = AWLocationNarrative(.CoordX * 1000, CellZ * 1000) & vbTab & " H=" & .Height
            If (.Texture = AW_TERRAIN_TEXTURE_HOLE) Then
              cellData = cellData & " [HOLE]"
              
            Else
              cellData = cellData & " T=" & (.Texture And &H3F)
              
              If ((.Texture And AW_TERRAIN_ROTATE_RIGHT) = AW_TERRAIN_ROTATE_RIGHT) Then
                cellData = cellData & " R=270"
              
              ElseIf ((.Texture And AW_TERRAIN_ROTATE_LEFT) = AW_TERRAIN_ROTATE_LEFT) Then
                cellData = cellData & " R=90"
                
              ElseIf ((.Texture And AW_TERRAIN_ROTATE_OPPOSITE) = AW_TERRAIN_ROTATE_OPPOSITE) Then
                cellData = cellData & " R=180"
              
              End If
            End If
            
            FormWriteToLog cellData
            terrainCount = terrainCount + 1
            
          End If
        End With
      Next indexX
    End With
  Next CellZ
  
  FormWriteToLog "## Total Markers: " & terrainCount & ", Total Cells: " & cellCount

End Sub


'***************************************************************************
'** Add a new item to the Show/Hide Queue                                 **
'***************************************************************************
Private Sub QueueShowHideRequest(ByVal terrainX As String, ByVal terrainZ As Long, ByVal markerVisible As Boolean)

  mShowHideQueue.Add CStr(terrainX) & " " & CStr(terrainZ) & " " & IIf(markerVisible, TERRAIN_MARKER_CMD_SHOW, TERRAIN_MARKER_CMD_HIDE)
  
End Sub


'***************************************************************************
'** Process Show/Hide Queue                                               **
'***************************************************************************
Public Sub TerrainShowHideQueueProcess()

  Dim queuedRequest         As String
  Dim terrainMarker         As STRUCT_TERRAIN_MARKER
  Dim markerVisible         As Boolean
  Dim cmdParms()            As String
  
  '*************************************************************************
  '** Do nothing if show/hide queue is empty                              **
  '*************************************************************************
  If Not (mShowHideQueue.Count = 0) Then
    
    '***********************************************************************
    '** Retrieve the cell details from the queue and remove the item      **
    '***********************************************************************
    queuedRequest = mShowHideQueue(1)
    cmdParms = Split(queuedRequest, " ")
    mShowHideQueue.Remove 1
    
    '***********************************************************************
    '** If the cell is previously unknown, query it                       **
    '***********************************************************************
    terrainMarker = TerrainCacheMarkerGet(Val(cmdParms(0)), Val(cmdParms(1)))
    If Not (terrainMarker.Number = TERRAIN_NO_MARKER) Then
      markerVisible = IIf(cmdParms(2) = TERRAIN_MARKER_CMD_SHOW, True, False)
      
      '*********************************************************************
      '** Do nothing if the marker's visibility is already as required    **
      '*********************************************************************
      If Not (terrainMarker.Visible = markerVisible) Then
      
        '*******************************************************************
        '** Update the object in the world                                **
        '*******************************************************************
        terrainMarker.Visible = markerVisible
        If Not (TerrainMarkerLoad(terrainMarker) = AW_RC_OK) Then
          mShowHideQueue.Add queuedRequest
          
        End If
      End If
    End If
  End If
  
End Sub


'***************************************************************************
'** Add a new cell to the terrain cache                                   **
'***************************************************************************
Private Function TerrainCacheAddNewCell(ByVal terrainCellX As Long, ByVal terrainCellZ As Long) As Long

  Static cellXIndex         As Long
  Static scanXIndex         As Long
  
  '*************************************************************************
  '** Default the index to "Not Found"                                    **
  '*************************************************************************
  cellXIndex = TERRAIN_NO_CELL_INDEX
  
  '*************************************************************************
  '** Scan the X cell list for the supplied Z cell for the X cell number  **
  '*************************************************************************
  With mTerrainCache(terrainCellZ)
    scanXIndex = 1
    
    Do While (scanXIndex <= .cellCount)
      
      '*********************************************************************
      '** keep scanning while the X cell has not been reached             **
      '*********************************************************************
      If (.XCells(scanXIndex).CoordX < terrainCellX) Then
        scanXIndex = scanXIndex + 1
        
      Else '.XCells(scanXIndex).CoordX >= terrainCellx
        cellXIndex = scanXIndex
        Exit Do
      
      End If
    Loop
    
    '***********************************************************************
    '** If we never reached the X cell, add a new cell to the end         **
    '***********************************************************************
    If (cellXIndex = TERRAIN_NO_CELL_INDEX) Then
      .cellCount = .cellCount + 1
      ReDim Preserve .XCells(.cellCount)
      cellXIndex = .cellCount
      
      With .XCells(cellXIndex)
        .CoordX = terrainCellX
        .Owner = TERRAIN_NO_MARKER
        .ObjectNumber = TERRAIN_NO_MARKER
        .Timestamp = TERRAIN_NO_MARKER
      End With
      
      mTerrainCellCount = mTerrainCellCount + 1
      FormSetCellCount mTerrainCellCount
      
    '***********************************************************************
    '** Do nothing if the cell is already known                           **
    '***********************************************************************
    ElseIf Not (.XCells(cellXIndex).CoordX = terrainCellX) Then
      
      '*********************************************************************
      '** Shuffle the array up one for cells after the scan cell          **
      '*********************************************************************
      .cellCount = .cellCount + 1
      ReDim Preserve .XCells(.cellCount)
    
      For scanXIndex = .cellCount - 1 To cellXIndex Step -1
        .XCells(scanXIndex + 1) = .XCells(scanXIndex)
        
      Next scanXIndex
        
      '*********************************************************************
      '** Set the details of the new cell to defaults                     **
      '*********************************************************************
      With .XCells(cellXIndex)
        .CoordX = terrainCellX
        .Owner = TERRAIN_NO_MARKER
        .ObjectNumber = TERRAIN_NO_MARKER
        .Timestamp = TERRAIN_NO_MARKER
      End With
    
      '*********************************************************************
      '** Increment the cache cell count                                  **
      '*********************************************************************
      mTerrainCellCount = mTerrainCellCount + 1
      FormSetCellCount mTerrainCellCount
    
    End If
  End With
  
  '*************************************************************************
  '** Return the index of the new or existing cell                        **
  '*************************************************************************
  TerrainCacheAddNewCell = cellXIndex
  
End Function


'***************************************************************************
'** Retrieve the array index of the X cell for the supplied Z cell        **
'***************************************************************************
Private Function TerrainCacheGetCellXIndex(ByVal terrainCellX As Long, ByVal terrainCellZ As Long) As Long

  Static cellXIndex         As Long
  Static scanXIndex         As Long
  
  Static rangeXMin          As Long
  Static rangeXMax          As Long
  Static scanComplete       As Boolean
  
  '*************************************************************************
  '** Default the index to "Not Found"                                    **
  '*************************************************************************
  cellXIndex = TERRAIN_NO_CELL_INDEX
  
  With mTerrainCache(terrainCellZ)
  
    '***********************************************************************
    '** If only a few elements exist in the list, just scan sequentially  **
    '***********************************************************************
    If (.cellCount < 15) Then
      For scanXIndex = 1 To .cellCount
        If (.XCells(scanXIndex).CoordX = terrainCellX) Then
          cellXIndex = scanXIndex
          Exit For
        
        ElseIf (.XCells(scanXIndex).CoordX > terrainCellX) Then
          Exit For
        
        End If
      Next scanXIndex
    
    Else
    
      '*********************************************************************
      '** Initialise the binary search                                    **
      '*********************************************************************
      rangeXMin = 1
      rangeXMax = .cellCount
      scanComplete = False
      
      '*********************************************************************
      '** Keep scanning until we get a result either way                  **
      '*********************************************************************
      Do While Not (scanComplete)
        scanXIndex = rangeXMin + Fix((rangeXMax - rangeXMin) / 2)
        
        '*********************************************************************
        '** If we can't halve the remaining array any more, comapre the     **
        '** min/max values for a match                                      **
        '*********************************************************************
        If (scanXIndex = rangeXMin) Then
          If (.XCells(rangeXMin).CoordX = terrainCellX) Then
            cellXIndex = rangeXMin
          
          ElseIf (.XCells(rangeXMax).CoordX = terrainCellX) Then
            cellXIndex = rangeXMax
            
          End If
          
          scanComplete = True
          Exit Do
          
        '*********************************************************************
        '** If the desired cell x coord is less than the midpoint, use the  **
        '** bottom half of the array and repeat                             **
        '*********************************************************************
        ElseIf (.XCells(scanXIndex).CoordX > terrainCellX) Then
          rangeXMax = scanXIndex
          
        '*********************************************************************
        '** If the desired cell x coord is greater than the midpoint, use   **
        '** the top half of the array and repeat                            **
        '*********************************************************************
        Else
          rangeXMin = scanXIndex
          
        End If
      Loop
    End If
  End With
    
  
  '*************************************************************************
  '** Return the index of the X cell in the list                          **
  '*************************************************************************
  TerrainCacheGetCellXIndex = cellXIndex
  
End Function

